library mvcprovider;

import 'package:flutter/material.dart';
import 'package:provider/provider.dart';

///
/// The module class which is the link between MVC entities
///
/// Sample :
///
/// class MyCounterModule extends Module<MyCounterModel, MyCounterView, MyCounterCtrl> {
///
///   final MyCounterModel model = MyCounterModel();
///   final MyCounterView view = MyCounterView();
///   final MyCounterCtrl ctrl = MyCounterCtrl();
/// }
class Module<M extends Model, V extends Widget, C extends Controller>
    extends StatelessWidget {
  final M model;
  final V view;
  final C ctrl;
  final Map<Type, dynamic> dependencies;

  Module({this.model, this.view, this.ctrl, this.dependencies});

  @override
  Widget build(BuildContext context) {
    return Container(
        child: MultiProvider(providers: [
          ChangeNotifierProvider(create: (_) => model),
          ChangeNotifierProvider(create: (_) => ctrl.init(_, model, dependencies))
        ], child: view));
  }
}


///
/// The Model class which will contain data and optionally
/// some methods to format this data
///
/// Sample :
///
/// class MyCounterModel extends Model {
///
///   int count = 0;
/// }
class Model extends ChangeNotifier {}

///
/// The Controller class which will handle user's actions
///
/// Sample :
///
/// class MyCounterCtrl extends Controller<MyCounterModel> {
///
///   void increment() {
///     model.count++;
///     notifyListeners();
///   }
/// }
class Controller<M> extends ChangeNotifier {
  BuildContext context;
  M model;
  Map<Type, dynamic> dependencies;

  Controller init(
      BuildContext context, M model, Map<Type, dynamic> dependencies) {
    this.context = context;
    this.model = model;
    this.dependencies =
        (dependencies ?? {}).map((type, fct) => MapEntry(type, fct(context)));
    return this;
  }

  T get<T>() => dependencies[T];
}

///
/// A private class which consist in a type definition
/// in order to keep view's properties immutable
///
class _MC<M, C> { M model; C ctrl; }

///
/// The View mixin which will display model data and link
/// controller's methods to user's actions
///
/// Sample :
///
/// class MyCounterView extends View<MyCounterModel, MyCounterCtrl> {
///
///   @override
///   Widget build(BuildContext context, [Widget child]) {
///     super.build(context);
///
///     return Column(
///         mainAxisAlignment: MainAxisAlignment.center,
///         children: <Widget>[
///           Text("Counter : ${model.count}"),
///           RaisedButton(
///             child: Text("Increment"),
///             onPressed: ctrl.increment,
///           )
///         ]
///     );
///   }
/// }
mixin View<M extends Model, C extends Controller> {
  final _MC<M, C> _ = new _MC();

  M get model => _.model;
  C get ctrl => _.ctrl;

  void init(BuildContext context) {
    _.model = Provider.of<M>(context);
    _.ctrl = Provider.of<Controller>(context);
  }
}
